import sys
import os
import numpy as np
import itertools as it
from collections import OrderedDict, defaultdict
from numpy import linalg as LA
import json
from ase.units import Bohr, Hartree


def _write(elem1, elem2, CCS_params, exp=True):
    elem1 = elem1
    elem2 = elem2
    no_pair = False
    try:
        pair = elem1 + "-" + elem2
        rcut = CCS_params["Two_body"][pair]["r_cut"]
    except:
        try:
            pair = elem2 + "-" + elem1
            rcut = CCS_params["Two_body"][pair]["r_cut"]
        except:
            rcut = 0.0
            no_pair = True
    if no_pair:
        pass
    else:
        Rmin = CCS_params["Two_body"][pair]["r_min"] / Bohr
        Rcut = CCS_params["Two_body"][pair]["r_cut"] / Bohr
        a = CCS_params["Two_body"][pair]["spl_a"]
        b = CCS_params["Two_body"][pair]["spl_b"]
        c = CCS_params["Two_body"][pair]["spl_c"]
        d = CCS_params["Two_body"][pair]["spl_d"]
        aa = CCS_params["Two_body"][pair]["exp_a"]
        bb = CCS_params["Two_body"][pair]["exp_b"]
        cc = CCS_params["Two_body"][pair]["exp_c"]
        exp = False
        x = CCS_params["Two_body"][pair]["r"]
        dx = CCS_params["Two_body"][pair]["dr"]

        with open(f"{elem1}-{elem2}.spl", "w") as f:
            print("Spline", file=f)
            print(len(x), (Rcut + dx) / Bohr, file=f)
            print(aa * Bohr, bb + np.log(1 / Hartree), cc / Hartree, file=f)
            for i in range(len(x)):
                if i < (len(x) - 1):
                    print(
                        x[i] / Bohr,
                        (x[i+1]) / Bohr,
                        a[i] / Hartree,
                        b[i] * (Bohr) / Hartree,
                        c[i] * (Bohr**2) / Hartree,
                        d[i] * (Bohr**3) / Hartree,
                        file=f,
                    )
                else:
                    print(
                        x[i] / Bohr,
                        (x[i] + dx) / Bohr,
                        a[i] / Hartree,
                        b[i] * (Bohr) / Hartree,
                        c[i] * (Bohr**2) / Hartree,
                        d[i] * (Bohr**3) / Hartree,
                        "0.0 0.0",
                        file=f,
                    )

        with open(f"{elem2}-{elem1}.spl", "w") as f:
            print("Spline", file=f)
            print(len(x), (Rcut + dx) / Bohr, file=f)
            print(aa * Bohr, bb + np.log(1 / Hartree), cc / Hartree, file=f)
            for i in range(len(x)):
                if i < (len(x) - 1):
                    print(
                        x[i] / Bohr,
                        (x[i+1]) / Bohr,
                        a[i] / Hartree,
                        b[i] * (Bohr) / Hartree,
                        c[i] * (Bohr**2) / Hartree,
                        d[i] * (Bohr**3) / Hartree,
                        file=f,
                    )
                else:
                    print(
                        x[i] / Bohr,
                        (x[i] + dx) / Bohr,
                        a[i] / Hartree,
                        b[i] * (Bohr) / Hartree,
                        c[i] * (Bohr**2) / Hartree,
                        d[i] * (Bohr**3) / Hartree,
                        "0.0 0.0",
                        file=f,
                    )


def write_dftb_spline(CCS_params_file):

    with open(CCS_params_file, "r") as f:
        CCS_params = json.load(f)

    for pair in CCS_params["Two_body"]:
        elem = pair.split("-")
        _write(elem[0], elem[1], CCS_params, exp=True)


def main():

    size = os.get_terminal_size()
    c = size.columns
    txt = "-"*c
    print("")
    print(txt)

    try:
        import art
        txt = art.text2art('CCS:export sk-table')
        print(txt)
    except:
        print('CCS: export sk-table')

    try:
        CCS_params_file = sys.argv[1]
    except:
        print("Please provide CCS params-file as first argument.")
        exit()

    size = os.get_terminal_size()
    c = size.columns
    txt = "-"*c
    print(txt)
    print("")

    write_dftb_spline(CCS_params_file)


if __name__ == "__main__":
    main()



import sys
import os
import numpy as np
import ase.db as db
import json
import random
from ase.calculators.singlepoint import SinglePointCalculator
from tqdm import tqdm

from ccs_fit.ase_calculator.ccs_ase_calculator import CCS


def ccs_validate(
    mode=None,
    CCS_params=None,
    Ns='all',
    DFT_DB="DFT.db",
    CCS_DB="CCS_validate.db",
    DFTB_DB=None,
    charge_dict=None,
    charge_scaling=False,
    include_forces=False,
    verbose=False,
):
    """
    Function to verify database generation.
    ---------------------------------------

    Input
    -----
        mode : str
            String describing which mode is used, supported modes are:
                CCS:   CCS_params_file(string) NumberOfSamples(int) DFT.db(string)")
                CCS+Q: CCS_params_file(string) NumberOfSamples(int) DFT.db(string) charge_dict(string) charge_scaling(bool)")
                DFTB:  CCS_params_file(string) NumberOfSamples(int) DFT.db(string) DFTB.db(string)")

    Returns
    -------
        What does it return Jolla?

    Example usage
    -------------
        ccs_validate MODE [...]
    """
    if os.path.isfile(CCS_DB):
        os.remove(CCS_DB)

    DFT_DB = db.connect(DFT_DB)
    CCS_DB = db.connect(CCS_DB)

    if mode == "DFTB":
        DFTB_DB = db.connect(DFTB_DB)

    if isinstance(charge_dict, str):
        charge_dict = json.loads(charge_dict)

    if isinstance(CCS_params, str):
        with open(CCS_params, "r") as f:
            CCS_params = json.load(f)

    if charge_dict is None:
        charge = False
    else:
        charge = True

    f = open("CCS_validate.dat", "w")
    print("{:^13s} {:^13s} {:^13s} {:^13s} {:^13s}".format("#Reference", "Predicted", "Error", "No_of_atoms", "structure_no"), file=f)

    if include_forces:
        f_force = open("CCS_validate_forces.dat", "w")
        print("{:^13s} {:^13s} {:^13s} {:^13s}".format("#Reference", "Predicted", "Error", "structure_no"), file=f_force)

    calc = CCS(CCS_params=CCS_params, charge=charge,
               q=charge_dict, charge_scaling=charge_scaling)

    if Ns == 'all':
        Ns = -1  # CONVERT TO INTEGER INPUT FORMAT

    if Ns > 0:
        Ns = int(Ns)
        mask = [a <= Ns for a in range(len(DFT_DB))]
        random.shuffle(mask)
    else:
        mask = len(DFT_DB) * [True]

    counter = 0
    for row in tqdm(DFT_DB.select(), total=len(DFT_DB), colour="#800000"):
        if mask[counter]:
            structure = row.toatoms()
            EDFT = structure.get_potential_energy()
            if include_forces:
                FREF = structure.get_forces()
            EREF = EDFT
            structure.calc = calc
            ECCS = structure.get_potential_energy()
            if include_forces:
                FCCS = structure.get_forces()
            if mode == "DFTB":
                key = row.key
                EDFTB = DFTB_DB.get("key=" + str(key)).energy
                EREF = EDFT - EDFTB
                if include_forces:
                    FDFTB= DFTB_DB.get("key=" + str(key)).forces
                    FREF = [FREF[i]-FDFTB[i] for i in range(len(FREF))]
                sp_calculator = SinglePointCalculator(
                    structure, energy=EDFTB + ECCS)
                structure.calc = sp_calculator
                structure.get_potential_energy()

            print('{:13.8f} {:13.8f} {:13.8f} {:13d} {:13d}'.format(
                    EREF, ECCS, np.abs(EREF - ECCS),
                    len(structure), counter), file=f)
            if include_forces:
                FREF = [item for sublist in FREF for item in sublist]
                FCCS = [item for sublist in FCCS for item in sublist]
                for force_id, force_ref in enumerate(FREF):
                    print('{:13.8f} {:13.8f} {:13.8f} {:13d}'.format(
                        force_ref, FCCS[force_id], np.abs(force_ref - FCCS[force_id]), counter), file=f_force)

            try:
                CCS_DB.write(structure, key=key)
            except:
                CCS_DB.write(structure)
        
        counter += 1


def main():
    import argparse

    try:
        size = os.get_terminal_size()
        c = size.columns
        txt = "-"*c
        print("")
        print(txt)
        import art
        txt = art.text2art('CCS:Validate')
        print(txt)
    except:
        pass

    parser = argparse.ArgumentParser(description='CCS fetching tool')
    parser.add_argument("-m", "--mode",         type=str, metavar="",
                        default='CCS',  help="Mode. Availble option: CCS, CCS+Q, DFTB")
    parser.add_argument("-d", "--DFT_DB", type=str, metavar="",
                        default='DFT.db',  help="Name of DFT reference data-base")
    parser.add_argument("-dc", "--CCS_DB", type=str, metavar="",
                        default='CCS_validate.db',  help="Name of data-base to store results")
    parser.add_argument("-p", "--CCS_params", type=str, metavar="",
                        default='CCS_params.json',  help="CCS_params.json file")
    parser.add_argument("-n", "--Ns",  type=int,  metavar="",
                        default=-1,  help="Number of structures to include")
    parser.add_argument("-v", "--verbose",
                        action="store_true", help="Verbose output")
    parser.add_argument("-chg", "--charge_dict",      type=json.loads, metavar="",
                        help="Specify atomic charges in json format, e.g.: \n \'{ \"Zn\" : 2.0 , \"O\" : -2.0 }\'  ")
    parser.add_argument("-chg_s", "--charge_scaling",
                        type=bool, metavar="", default=False)
    parser.add_argument("-f", "--include_forces", type=bool, metavar="",
                        default=False, help="Validation of the reproduced forces w.r.t. those they were fitted on.")

    args = parser.parse_args()

    ccs_validate(**vars(args))


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------#
#  CCS: Curvature Constrained Splines                                          #
#  Copyright (C) 2019 - 2023  CCS developers group                             #
#                                                                              #
#  See the LICENSE file for terms of usage and distribution.                   #
# ------------------------------------------------------------------------------#

import numpy as np
from ctypes import *
from ppmd import *
import json
from ppmd.coulomb.fmm import internal_to_ev
import math
from scipy.optimize import minimize as min
from time import time as time
from ase.calculators.singlepoint import SinglePointCalculator


# CLASSES AND FUNCTION TO EXPLOIT THE FAST ELECTROSTATIC SOLVERS OF THE PPMD CODE


# some alias for readability and easy modification if we ever
# wish to use CUDA.

PairLoop = pairloop.CellByCellOMP
ParticleLoop = loop.ParticleLoopOMP
State = state.State
PositionDat = data.PositionDat
ParticleDat = data.ParticleDat
ScalarArray = data.ScalarArray
Kernel = kernel.Kernel
GlobalArray = data.GlobalArray
Constant = kernel.Constant
IntegratorRange = method.IntegratorRange


def steepest_descent(A, constants):

    sd_kernel_code = """
    P.i[0] += alpha*F.i[0];
    P.i[1] += alpha*F.i[1];
    P.i[2] += alpha*F.i[2];
    """
    sd_kernel = Kernel("SD", sd_kernel_code, constants)
    SD = ParticleLoop(
        kernel=sd_kernel, dat_dict={"F": A.force(access.READ), "P": A.pos(access.INC)}
    )

    return SD


def newtonic(A, constants):

    N_kernel_code = """
    double dx = 4.0*abs((P.i[0]- P_old.i[0])/(F_old.i[0]-F.i[0]));
    double dy = 4.0*abs((P.i[1]- P_old.i[1])/(F_old.i[1]-F.i[1]));
    double dz = 4.0*abs((P.i[2]- P_old.i[2])/(F_old.i[2]-F.i[2]));

    P_old.i[0]=P.i[0];
    P_old.i[1]=P.i[1];
    P_old.i[2]=P.i[2];

    P.i[0] += (dx < alpha) ? F.i[0]*dx: alpha*F.i[0];
    P.i[1] += (dy < alpha) ? F.i[1]*dy: alpha*F.i[1];
    P.i[2] += (dz < alpha) ? F.i[2]*dz: alpha*F.i[2];

    F_old.i[0]=F.i[0];
    F_old.i[1]=F.i[1];
    F_old.i[2]=F.i[2];
    """
    N_kernel = Kernel("NK", N_kernel_code, constants)
    NK = ParticleLoop(
        kernel=N_kernel,
        dat_dict={
            "F": A.force(access.READ),
            "P": A.pos(access.INC),
            "F_old": A.force_old(access.INC),
            "P_old": A.pos_old(access.INC),
        },
    )

    return NK


class FIRE:
    def __init__(self, A, CCS=None, dt_start=0.1, a_start=0.1, dt_max=1.0):
        self.CCS = CCS
        self.dt_start = dt_start
        self.a_start = a_start
        self.constants = (Constant("dr_max", 0.2),)
        self.nmin = 5
        self.f_inc = 1.1
        self.f_dec = 0.5
        self.f_a = 0.99
        self.dt_max = dt_max
        A.F_sq = GlobalArray(ncomp=1, dtype=c_double)
        A.V_sq = GlobalArray(ncomp=1, dtype=c_double)
        A.a = GlobalArray(ncomp=1, dtype=c_double)
        A.dt = GlobalArray(ncomp=1, dtype=c_double)
        A.dr_norm = GlobalArray(ncomp=1, dtype=c_double)
        A.Fv = GlobalArray(ncomp=1, dtype=c_double)
        A.dt.set(dt_start)
        A.a.set(a_start)

        self.F1 = self.FIRE_LOOP_1(A)
        self.F2 = self.FIRE_LOOP_2(A)
        self.F3 = self.FIRE_LOOP_3(A)
        self.F4 = self.FIRE_LOOP_4(A)
        self.F5 = self.FIRE_LOOP_5(A)

    def FIRE_LOOP_1(self, A):

        kernel_code = """
        P_old.i[0]=P.i[0];
        P_old.i[1]=P.i[1];
        P_old.i[2]=P.i[2];
        V_old.i[0]=V.i[0];
        V_old.i[1]=V.i[1];
        V_old.i[2]=V.i[2];
        F_old.i[0]=F.i[0];
        F_old.i[1]=F.i[1];
        F_old.i[2]=F.i[2];
        F_sq[0] += (F.i[0]*F.i[0] + F.i[1]*F.i[1] + F.i[2]*F.i[2]);
        """
        kernel = Kernel("k", kernel_code, self.constants)
        pl = ParticleLoop(
            kernel=kernel,
            dat_dict={
                "P": A.pos(access.READ),
                "P_old": A.pos_old(access.INC),
                "V": A.vel(access.READ),
                "V_old": A.vel_old(access.INC),
                "F": A.force(access.READ),
                "F_old": A.force_old(access.INC),
                "dt": A.dt(access.READ),
                "F_sq": A.F_sq(access.INC_ZERO),
            },
        )

        return pl

    def FIRE_LOOP_2(self, A):

        kernel_code = """
        const double norm = sqrt(F_sq[0]);
        F_norm.i[0] = F.i[0]/norm; 
        F_norm.i[1] = F.i[1]/norm; 
        F_norm.i[2] = F.i[2]/norm; 
        V_sq[0] += (V.i[0]*V.i[0] + V.i[1]*V.i[1] + V.i[2]*V.i[2]);
        Fv[0] += (F.i[0]*V.i[0] + F.i[1]*V.i[1] + F.i[2]*V.i[2])*14.39964547842567;
        """
        kernel = Kernel("k", kernel_code, self.constants)
        pl = ParticleLoop(
            kernel=kernel,
            dat_dict={
                "V": A.vel(access.INC),
                "F_norm": A.force_norm(access.INC_ZERO),
                "F": A.force(access.READ),
                "M": A.mass(access.READ),
                "F_sq": A.F_sq(access.READ),
                "dt": A.dt(access.READ),
                "Fv": A.Fv(access.INC_ZERO),
                "V_sq": A.V_sq(access.INC_ZERO),
            },
        )

        return pl

    def FIRE_LOOP_3(self, A):

        kernel_code = """
        const double norm = sqrt(V_sq[0]);
        V.i[0] = V.i[0]*(1.0-a[0]) + a[0]*F_norm.i[0]*norm;
        V.i[1] = V.i[1]*(1.0-a[0]) + a[0]*F_norm.i[1]*norm;
        V.i[2] = V.i[2]*(1.0-a[0]) + a[0]*F_norm.i[2]*norm;
        """
        kernel = Kernel("k", kernel_code, self.constants)
        pl = ParticleLoop(
            kernel=kernel,
            dat_dict={
                "V": A.vel(access.INC),
                "P": A.pos(access.INC),
                "F_norm": A.force_norm(access.READ),
                "F": A.force(access.READ),
                "dt": A.dt(access.READ),
                "M": A.mass(access.READ),
                "a": A.a(access.READ),
                "V_sq": A.V_sq(access.READ),
            },
        )

        return pl

    def FIRE_LOOP_4(self, A):

        kernel_code = """
        V.i[0] += dt[0]*F.i[0]*14.39964547842567;
        V.i[1] += dt[0]*F.i[1]*14.39964547842567;
        V.i[2] += dt[0]*F.i[2]*14.39964547842567;
        dr.i[0] = dt[0]*V.i[0];
        dr.i[1] = dt[0]*V.i[1];
        dr.i[2] = dt[0]*V.i[2];
        dr_norm[0] += (dr.i[0]*dr.i[0]+dr.i[1]*dr.i[1]+dr.i[2]*dr.i[2]);
        """
        kernel = Kernel("k", kernel_code, self.constants)
        pl = ParticleLoop(
            kernel=kernel,
            dat_dict={
                "V": A.vel(access.INC),
                "P": A.pos(access.INC),
                "dr": A.dr(access.INC),
                "dr_norm": A.dr_norm(access.INC_ZERO),
                "F_norm": A.force_norm(access.READ),
                "F": A.force(access.READ),
                "dt": A.dt(access.READ),
                "M": A.mass(access.READ),
                "a": A.a(access.READ),
                "V_sq": A.V_sq(access.READ),
            },
        )

        return pl

    def FIRE_LOOP_5(self, A):

        kernel_code = """
        const double norm = sqrt(dr_norm[0]);
        if(norm < dr_max){
        P.i[0] += dr.i[0];
        P.i[1] += dr.i[1];
        P.i[2] += dr.i[2];
        }
        if(norm >= dr_max){
        P.i[0] += dr_max*dr.i[0]/norm;
        P.i[1] += dr_max*dr.i[1]/norm;
        P.i[2] += dr_max*dr.i[2]/norm;
        }

        """
        kernel = Kernel("k", kernel_code, self.constants)
        pl = ParticleLoop(
            kernel=kernel,
            dat_dict={
                "V": A.vel(access.INC),
                "P": A.pos(access.INC),
                "dr": A.dr(access.READ),
                "F_norm": A.force_norm(access.READ),
                "F": A.force(access.READ),
                "dt": A.dt(access.READ),
                "M": A.mass(access.READ),
                "a": A.a(access.READ),
                "dr_norm": A.dr_norm(access.READ),
                "V_sq": A.V_sq(access.READ),
            },
        )

        return pl

    def opt(self, A, steps=10000, fmax=0.05):
        mem = 0
        Eold, _ = self.CCS.eval(A)
        t1 = time()
        for it in range(steps):
            uphill = False
            Enew, _ = self.CCS.eval(A=A)
            if Enew > Eold:
                A.pos[:] = A.pos_old[:]
                A.vel[:] = A.vel_old[:]
                A.force[:] = A.force_old[:]
                Enew = Eold
                uphill = True
            Eold = Enew
            self.F1.execute()
            self.F2.execute()
            if (A.Fv[0] > 0 and not uphill) or (it == 0):
                self.F3.execute()
                if mem > self.nmin:
                    A.dt.set(min(A.dt[0] * self.f_inc, self.dt_max))
                    A.a.set(A.a[0] * self.f_a)
                mem += 1
            else:
                mem = 0
                with A.vel.modify_view() as m:
                    m[:, :] = 0.0
                A.a.set(self.a_start)
                A.dt.set(A.dt[0] * self.f_dec)
            if not uphill:
                self.F4.execute()
                self.F5.execute()
            MaxF = (
                np.max(np.linalg.norm(A.force[0 : A.npart], axis=1)) * internal_to_ev()
            )
            if A.domain.comm.rank == 0 and it % 1 == 0:
                print(
                    "{: 8d} | E: {: 12.8f} F: {: 12.8f}  a: {: 12.8f} dt: {: 12.8g}  Fv: {: 12.8f} ".format(
                        it, Eold, MaxF, A.a[0], A.dt[0], A.Fv[0]
                    )
                )
            if MaxF < fmax:
                t2 = time()
                print(
                    "------------------------------- OPTIMIZATION COMPLETE -------------------------------"
                )
                print(
                    "    Total time for opimization: ",
                    t2 - t1,
                    " s, Max force:",
                    np.max(np.linalg.norm(A.force.view * internal_to_ev(), axis=1)),
                    "ev/Å",
                )
                break


class CCS:
    def __init__(self, A, atoms, q=None, CCS_params="CCS_params.json"):

        A.Etot = GlobalArray(ncomp=1, dtype=c_double)

        with open(CCS_params, "r") as f:
            CCS_params = json.load(f)

        id = {}
        charges = np.zeros((len(atoms), 1))
        at_ids = np.zeros((len(atoms), 1))
        keys = list(q.keys())
        Ntypes = len(keys)

        for a in keys:
            id[a] = keys.index(a)

        one_body_E = 0.0
        i = -1
        for a in atoms.get_chemical_symbols():
            i += 1
            charges[i] = q[a] * CCS_params["Charge scaling factor"]
            at_ids[i] = id[a]
            try:
                one_body_E += CCS_params["One_body"][a] / 14.39964547842567
            except:
                pass

        N = len(atoms)
        Ex = atoms.cell[0, 0]
        Ey = atoms.cell[1, 1]
        Ez = atoms.cell[2, 2]

        # make a state object and set the global number of particles N

        A.npart = N
        A.Ex = Ex
        A.Ey = Ey
        A.Ez = Ez

        # give the state a domain and boundary condition
        A.domain = domain.BaseDomainHalo(extent=(Ex, Ey, Ez))
        A.domain.boundary_condition = domain.BoundaryTypePeriodic()

        # add a PositionDat to contain positions
        A.pos = PositionDat(ncomp=3)
        A.force = ParticleDat(ncomp=3)
        A.charge = ParticleDat(ncomp=1)
        A.at = ParticleDat(ncomp=1)  # ,dtype= c_int)
        A.mass = ParticleDat(ncomp=1)
        A.vel = ParticleDat(ncomp=3)

        # add spline table
        SP_interval = ScalarArray(ncomp=Ntypes**2, dtype=c_int)
        SP_Rmin = ScalarArray(ncomp=Ntypes**2)
        SP_Rcut = ScalarArray(ncomp=Ntypes**2)
        SP_Ntypes = ScalarArray(ncomp=1, dtype=c_int)
        SP_Ntypes[:] = Ntypes

        table_lenght = 0
        cnt = -1
        for Z_i in id:
            for Z_j in id:
                cnt += 1
                try:
                    try:
                        tmp_Rcut = CCS_params["Two_body"][Z_i + "-" + Z_j]["r_cut"]
                        tmp_Rmin = CCS_params["Two_body"][Z_i + "-" + Z_j]["r_min"]
                        tmp_A = CCS_params["Two_body"][Z_i + "-" + Z_j]["spl_a"]
                    except:
                        tmp_Rcut = CCS_params["Two_body"][Z_j + "-" + Z_i]["r_cut"]
                        tmp_Rmin = CCS_params["Two_body"][Z_j + "-" + Z_i]["r_min"]
                        tmp_A = CCS_params["Two_body"][Z_j + "-" + Z_i]["spl_a"]
                except:
                    tmp_Rcut = 0.0
                    tmp_Rmin = 0.0
                    tmp_A = [0.0]

                k = len(tmp_A)
                table_lenght += k
                SP_Rmin[cnt] = tmp_Rmin
                SP_Rcut[cnt] = tmp_Rcut
                SP_interval[cnt] = table_lenght

        SP_A = ScalarArray(ncomp=table_lenght)
        SP_B = ScalarArray(ncomp=table_lenght)
        SP_C = ScalarArray(ncomp=table_lenght)
        SP_D = ScalarArray(ncomp=table_lenght)

        table_lenght = 0  # THIS TIME FILL IN THE ACTUAL TABLES...
        cnt = -1
        for Z_i in id:
            for Z_j in id:
                cnt += 1
                try:
                    try:
                        tmp_A = CCS_params["Two_body"][Z_i + "-" + Z_j]["spl_a"]
                        tmp_B = CCS_params["Two_body"][Z_i + "-" + Z_j]["spl_b"]
                        tmp_C = CCS_params["Two_body"][Z_i + "-" + Z_j]["spl_c"]
                        tmp_D = CCS_params["Two_body"][Z_i + "-" + Z_j]["spl_d"]
                    except:
                        tmp_A = CCS_params["Two_body"][Z_j + "-" + Z_i]["spl_a"]
                        tmp_B = CCS_params["Two_body"][Z_j + "-" + Z_i]["spl_b"]
                        tmp_C = CCS_params["Two_body"][Z_j + "-" + Z_i]["spl_c"]
                        tmp_D = CCS_params["Two_body"][Z_j + "-" + Z_i]["spl_d"]
                except:
                    tmp_A = [0.0]
                    tmp_B = [0.0]
                    tmp_C = [0.0]
                    tmp_D = [0.0]

                k = len(tmp_A)
                table_lenght += k
                SP_A[table_lenght - k : table_lenght] = tmp_A
                SP_B[table_lenght - k : table_lenght] = tmp_B
                SP_C[table_lenght - k : table_lenght] = tmp_C
                SP_D[table_lenght - k : table_lenght] = tmp_D

        # Fill in A (atoms) and put origin is in the centre

        pos = (atoms.get_scaled_positions() - [0.5, 0.5, 0.5]) * [Ex, Ey, Ez]
        A.pos[:] = pos
        A.force[:] = 0.0
        A.at[:] = at_ids
        A.charge[:] = charges
        A.mass[:] = atoms.get_masses()[:][0]
        A.pos_old = ParticleDat(ncomp=3)
        A.force_old = ParticleDat(ncomp=3)
        A.pos_old[:] = A.pos[:]
        A.force_old[:] = A.force[:]
        A.vel_old = ParticleDat(ncomp=3)
        A.dr = ParticleDat(ncomp=3)
        A.force_norm = ParticleDat(ncomp=3)

        # broadcast the data accross MPI ranks
        A.scatter_data_from(0)

        # the pairloop guarantees that all particles such that |r_i - r_j| < r_n
        # are looped over. It may also propose pairs of particles such that
        # |r_i - r_j| >= r_n and it is the users responsibility to mask off these
        # cases
        kernel_src = """
        // Vector displacement from particle i to particle j.
        const double R0 = P.j[0] - P.i[0];
        const double R1 = P.j[1] - P.i[1];
        const double R2 = P.j[2] - P.i[2];

        // Distance, dr and spline interval ...
        const int pi = (int)at.i[0]*(int)SP_Ntypes[0] + (int)at.j[0];
        const double r2 = R0*R0 + R1*R1 + R2*R2;
        const double r  = sqrt(r2);


        if(r < SP_Rcut[pi]){
        const int k_end = SP_interval[pi]-1;
        int k_start = 0;
        if(pi > 0){
            k_start = SP_interval[pi-1];
        } 

        const double delta = (SP_Rcut[pi]-SP_Rmin[pi] ) / (k_end-k_start);
        const int r_int = floor( (r-SP_Rmin[pi])/delta);
        const double dr= r - SP_Rmin[pi] - r_int*delta;
        const double E = SP_A[r_int+k_start]+dr*(SP_B[r_int+k_start] +dr*(SP_C[r_int+k_start] + dr*SP_D[r_int+k_start]));
        const double f= (1/14.39964547842567)*(SP_B[r_int+k_start] +2*dr*(SP_C[r_int+k_start] + 1.5*dr*SP_D[r_int+k_start]));
        U[0] += 0.5*E / 14.39964547842567; // Complying with ppmd internal units
        F.i[0] += f*R0/r ;
        F.i[1] += f*R1/r ;
        F.i[2] += f*R2/r ;
        }

        """

        ccs_kernel = Kernel("CCS", kernel_src)

        # create a pairloop

        r_c = max(SP_Rcut[:])  # Global cutoff radius
        r_n = r_c + 0.1 * r_c

        ccs_pairloop = PairLoop(
            ccs_kernel,
            {
                "P": A.pos(access.READ),
                "F": A.force(access.INC_ZERO),
                "U": A.Etot(access.INC_ZERO),
                "SP_A": SP_A(access.READ),
                "SP_B": SP_B(access.READ),
                "SP_C": SP_C(access.READ),
                "SP_D": SP_D(access.READ),
                "SP_interval": SP_interval(access.READ),
                "SP_Rmin": SP_Rmin(access.READ),
                "SP_Rcut": SP_Rcut(access.READ),
                "SP_Ntypes": SP_Ntypes(access.READ),
                "at": A.at(access.READ),
            },
            shell_cutoff=r_n,
        )
        l = 12

        if np.linalg.norm(charges) > 0.0001:
            try:
                es = coulomb.fmm.PyFMM(
                    A.domain, free_space=False, r=max(3, math.log(N, 8)), l=l
                )
            except:
                es = coulomb.ewald.EwaldOrthoganal(A.domain, real_cutoff=10.0)
            self.es = es
        else:
            self.es = None
            print("SWITHING OF +Q.")

        self.twobody = ccs_pairloop
        self.onebody = one_body_E

    def eval(self, A, pos=None):

        # if pos is None:
        #     pos=copy( A.pos.view)

        try:
            # print("POSITIONS SET SUCCESFULLY")
            pos = pos.reshape((A.npart, 3))
            pos[:, 0] = np.mod(pos[:, 0] + A.Ex, A.Ex) - 0.5 * A.Ex
            pos[:, 1] = np.mod(pos[:, 1] + A.Ey, A.Ey) - 0.5 * A.Ey
            pos[:, 2] = np.mod(pos[:, 2] + A.Ez, A.Ez) - 0.5 * A.Ez
            with A.pos.modify_view() as m:
                m[:, :] = pos
        except:
            # print("FIALED TO SET POSITIONS!")
            pass

        self.twobody.execute()
        E = A.Etot[0]
        if self.es:
            Ee = self.es(A.pos, A.charge, A.force)
            E += Ee
        return (
            E + self.onebody
        ) * internal_to_ev(), -A.force.view.flatten() * internal_to_ev()


def optimize(A, atoms, CCS, Type="LBGS", Fmax=0.05):
    def _fun(x, CCS, A):
        f, g = CCS.eval(A, pos=x)
        return f, g

    with A.pos.modify_view() as m:
        m[:, :] = atoms.positions

    if Type == "LBGS":
        options = {
            "disp": 1,
            "iprint": 101,
            "maxiter": 1000,
            "gtol": Fmax,
            "ftol": 1e-12,
        }
        method = "L-BFGS-B"

    if Type == "CG":
        options = {
            "disp": 1,
            "iprint": 101,
            "maxiter": 1000,
            "gtol": Fmax,
            "norm": np.inf,
        }
        method = "cg"

    t1 = time()
    min(
        _fun,
        A.pos[0 : A.npart].flatten(),
        args=(CCS, A),
        method=method,
        jac=True,
        options=options,
    )
    t2 = time()
    print(
        "------------------------------- OPTIMIZATION COMPLETE -------------------------------"
    )
    print(
        "    Total time for opimization: ",
        t2 - t1,
        " s, Max force:",
        np.max(np.linalg.norm(A.force.view * internal_to_ev(), axis=1)),
        "ev/Å",
    )
    atoms.positions = A.pos[0 : A.npart]
    calculator = SinglePointCalculator(
        atoms, energy=A.Etot[0], free_energy=A.Etot[0], forces=A.force[0 : A.npart]
    )

    atoms.calc = calculator
    atoms.get_potential_energy()
    atoms.get_forces()
    return atoms



import sys
import os
import numpy as np
import itertools as it
from collections import OrderedDict, defaultdict
from numpy import linalg as LA
import json
from ase.units import Bohr, Hartree
from scipy.optimize import curve_fit
from scipy.constants import epsilon_0
import matplotlib.pyplot as plt

def Buckingham(r, A, B, C):
    return A*np.exp(-B*r) - C/(r**6)

# def Buckingham_Coulomb(r, q1, q2, A, B, C):
#     return A*np.exp(-B*r) - C/(r**6) + q1*q2/(4*np.pi*epsilon_0*r)

def Lennard_Jones(r, eps, sigma):
    sig_r6 = (sigma/r)**6
    return 4*eps*(sig_r6**2 - sig_r6)

def Morse(r, De, a, re, C):
    return De*(1-np.exp(-a*(r-re)))**2 + C 

def _write(elem1, elem2, CCS_params, f_Buck, f_LJ, f_Mor, exp=True):
    elem1 = elem1
    elem2 = elem2
    no_pair = False
    try:
        pair = elem1 + "-" + elem2
        rcut = CCS_params["Two_body"][pair]["r_cut"]
    except:
        try:
            pair = elem2 + "-" + elem1
            rcut = CCS_params["Two_body"][pair]["r_cut"]
        except:
            rcut = 0.0
            no_pair = True
    if no_pair:
        pass
    else:
        Rmin = CCS_params["Two_body"][pair]["r_min"] 
        Rcut = CCS_params["Two_body"][pair]["r_cut"] 
        a = CCS_params["Two_body"][pair]["spl_a"]
        b = CCS_params["Two_body"][pair]["spl_b"]
        c = CCS_params["Two_body"][pair]["spl_c"]
        d = CCS_params["Two_body"][pair]["spl_d"]
        aa = CCS_params["Two_body"][pair]["exp_a"]
        bb = CCS_params["Two_body"][pair]["exp_b"]
        cc = CCS_params["Two_body"][pair]["exp_c"]
        exp = False
        x = CCS_params["Two_body"][pair]["r"]
        dx = CCS_params["Two_body"][pair]["dr"]

        r = np.linspace(Rmin, Rcut, 1000)
        spl_to_fit = []

        for cur_r in r:
            if cur_r >= Rmin and cur_r < Rcut:
                index = int(np.floor((cur_r - Rmin) / dx))
                dr = cur_r - x[index]
                f0 = a[index] + dr * (
                b[index] + dr * (c[index] + (d[index] * dr))
                )
                spl_to_fit.append(f0)
            elif cur_r < Rmin:
                val = np.exp(-aa * cur_r + bb) + cc
                spl_to_fit.append(val)
            elif cur_r >= Rcut:
                spl_to_fit.append(0)
    
        def Lennard_Jones(r, eps, sigma):
            sig_r6 = (sigma/r)**6
            return 4*eps*(sig_r6**2 - sig_r6)

        def Morse(r, De, a, re, C):
            return De*(1-np.exp(-a*(r-re)))**2 + C 

        popt_Buck, pcov = curve_fit(Buckingham, r, spl_to_fit, maxfev=5000)
        popt_Mor, pcov = curve_fit(Morse, r, spl_to_fit, p0=[2,1,1,0], bounds=[[0, 0, 0, -np.inf],[np.inf, np.inf, np.inf, np.inf]], maxfev=5000)
        popt_LJ, pcov = curve_fit(Lennard_Jones, r, spl_to_fit, p0=[1,1], maxfev=5000)
        print("Buckingham fit (not optimised) for element pair {}-{};     V(r) = {:.2f}*exp(-{:.2f}*r) -({:.2f})/r^6.".format(elem1, elem2, popt_Buck[0], popt_Buck[1], popt_Buck[2]))
        print("Lennard Jones fit (not optimised) for element pair {}-{};  V(r) = 4*{:.2f}*(({:.2f}/r)^12 - ({:.2f}/r)^6)".format(elem1, elem2, popt_LJ[0], popt_LJ[1], popt_LJ[1]))
        print("Morse fit (not optimised) for element pair {}-{};          V(r) = {:.2f}*(1-np.exp(-{:.2f}*(r-{:.2f})))^2 + {:.2f}".format(elem1, elem2, popt_Mor[0], popt_Mor[1], popt_Mor[2], popt_Mor[3]))


        print("{:^8s} {:^8s} {:20.10f} {:20.10f} {:20.10f}".format(elem1, elem2, popt_Buck[0], popt_Buck[1], popt_Buck[2]), file=f_Buck)
        print("{:^8s} {:^8s} {:20.10f} {:20.10f}".format(elem1, elem2, popt_LJ[0], popt_LJ[1]), file=f_LJ)
        print("{:^8s} {:^8s} {:20.10f} {:20.10f} {:20.10f} {:20.10f}".format(elem1, elem2, popt_Mor[0], popt_Mor[1], popt_Mor[2], popt_Mor[3]), file=f_Mor)


        plt.plot(r, spl_to_fit, 'r', label='CCS')
        plt.plot(r, Buckingham(r, popt_Buck[0], popt_Buck[1], popt_Buck[2]), 'b', label = 'Buck')
        plt.plot(r, Morse(r, popt_Mor[0], popt_Mor[1], popt_Mor[2], popt_Mor[3]), color='orange', label='Morse')
        plt.plot(r, Lennard_Jones(r, popt_LJ[0], popt_LJ[1]), color='magenta', label='LJ')
        plt.xlabel("Interatomic distance (Å)")
        plt.ylabel("Potential energy (eV)")
        plt.legend()
        plt.show()

def write_FF(CCS_params_file):

    with open(CCS_params_file, "r") as f:
        CCS_params = json.load(f)

    f_Buck = open("Buckingham.dat", "w")
    f_LJ = open("Lennard_Jones.dat", "w")
    f_Mor = open("Morse.dat", "w")

    print("{:^8s} {:^8s} {:^20s} {:^20s} {:^20s}\n".format("Element", "Element", "A", "B", "C"), file=f_Buck)
    print("{:^8s} {:^8s} {:^20s} {:^20s}\n".format("Element", "Element", "ε", "σ"), file=f_LJ)
    print("{:^8s} {:^8s} {:^20s} {:^20s} {:^20s} {:^20s}\n".format("Element", "Element", "D_e", "a", "r_e", "C"), file=f_Mor)
    
    for pair in CCS_params["Two_body"]:
        elem = pair.split("-")
        _write(elem[0], elem[1], CCS_params, exp=True, f_Buck=f_Buck, f_LJ=f_LJ, f_Mor=f_Mor)


def main():

    size = os.get_terminal_size()
    c = size.columns
    txt = "-"*c
    print("")
    print(txt)

    try:
        import art
        txt = art.text2art('Exporting Force Field parameters for Buckingham, Morse and Lennard Jones potentials.')
        print(txt)
    except:
        print('Exporting Force Field parameters for Buckingham, Morse and Lennard Jones potentials.')

    try:
        CCS_params_file = sys.argv[1]
    except:
        print("Please provide CCS params-file as first argument.")
        exit()

    size = os.get_terminal_size()
    c = size.columns
    txt = "-"*c
    print(txt)
    print("")

    write_FF(CCS_params_file)


if __name__ == "__main__":
    main()



#------------------------------------------------------------------------------#
#  CCS: Curvature Constrained Splines                                          #
#  Copyright (C) 2019 - 2023  CCS developers group                             #
#                                                                              #
#  See the LICENSE file for terms of usage and distribution.                   #
#------------------------------------------------------------------------------#

'''
Common Ewald summation tools used by the CCS project.
'''


from pymatgen import Lattice, Structure
from pymatgen.analysis import ewald


def ewald_summation(atoms):
    '''Calculates the Ewald summation of multiple structures.

    Args:

        atoms (list): list of ASE Atoms objects

    '''

    lattice = Lattice(atoms.get_cell())
    coords = atoms.get_scaled_positions()
    struct = Structure(lattice, atoms.get_chemical_symbols(), coords)
    struct.add_oxidation_state_by_guess()
    ew = ewald.EwaldSummation(struct, compute_forces=False)

    return ew.total_energy



import ase.db as db
import re
from ase.io import Trajectory, read, write
from ase.calculators.neighborlist import *
from ase.calculators.singlepoint import SinglePointCalculator
import numpy as np
from sympy import E
from tqdm import tqdm
import copy
import sys
from ase.units import Bohr, Hartree

try:
    from fortformat import Fnetdata
except:
    pass


def ccs_build_db(
    mode=None,
    DFT_DB=None,
    DFTB_DB=None,
    file_list=None,
    Fortnet=False,
    greedy=False,
    greed_threshold=0.0001,
):

    AUtoEvA = Hartree / Bohr

    DFT_DB = db.connect(DFT_DB)
    if mode == "DFTB":
        DFTB_DB = db.connect(DFTB_DB)

    f = open(file_list, "r")
    L = len(f.readlines())
    f.close()
    f = open(file_list, "r")

    fnet_energies = np.empty((L, 1))
    fnet_strucs = []
    fnet_features = []

    counter = 0
    for lns in tqdm(f, total=L):
        counter += 1
        lns = lns.split()
        DFT_FOLDER = lns[0]

        structure_DFT = read(DFT_FOLDER + "/OUTCAR", index=-1)
        EDFT = structure_DFT.get_potential_energy()
        DFT_DB.write(structure_DFT, PBE=True, key=counter)

        # EXTRACT ALL REASONABLE STEPS?
        converged_indices = []
        Natoms = len(structure_DFT)
        if greedy and (mode != "DFTB"):
            f2 = open(DFT_FOLDER + "/OUTCAR", "r")
            outcar = f2.read()
            f2.close
            NELM = int(re.findall("NELM   \=(.+?)\;", outcar)[0])
            indices = re.findall("Iteration(.+?)\(", outcar)
            all_energies = re.findall("entropy\=(.+?)e", outcar)
            all_energies = [float(x) / Natoms for x in all_energies]
            indices = [int(x) - 1 for x in indices]
            dE = re.findall("2. order\) :(.+?)\(", outcar)
            uindices = set(indices)
            converged_indices = []
            previous_E = EDFT
            for i in uindices:
                N_SCF = [(el == i) for el in indices]
                if (sum(N_SCF) < NELM) & (
                    abs(all_energies[i] - previous_E) > greed_threshold
                ):
                    converged_indices.append(i)
                    previous_E = all_energies[i]

            for i in tqdm(converged_indices):
                counter += 1
                structure_DFT = read(DFT_FOLDER + "/OUTCAR", index=i)
                EDFT = structure_DFT.get_potential_energy()
                DFT_DB.write(structure_DFT, PBE=True, key=counter)

        if mode == "DFTB":
            DFTB_FOLDER = lns[1]
            # READ DFTB
            structure_DFTB = copy.deepcopy(structure_DFT)
            f2 = open(DFTB_FOLDER + "/detailed.out", "r")
            EDFTBstr = f2.read()
            EDFTBstr = re.search("Total Electronic energy:(.+?)eV", EDFTBstr).group(1)
            EDFTBstr_arr = EDFTBstr.split()
            EDFTB = float(EDFTBstr_arr[2])
            f2.close

            # READ DFTB FORCES
            f2 = open(DFTB_FOLDER + "/detailed.out", "r")
            time_to_read = False
            DFTB_forces = np.zeros([structure_DFTB.get_global_number_of_atoms(), 3])
            while True:
                next_line = f2.readline()

                if next_line == " \n":
                    time_to_read = False

                if time_to_read:
                    af = next_line.split()
                    DFTB_forces[acnt, 0] = float(af[0]) * AUtoEvA
                    DFTB_forces[acnt, 1] = float(af[1]) * AUtoEvA
                    DFTB_forces[acnt, 2] = float(af[2]) * AUtoEvA
                    acnt += 1

                if "Total Forces" in next_line:
                    time_to_read = True
                    acnt = 0

                if not next_line:
                    break
            f2.close

            # READ MULLIKEN POPULATIONS
            f2 = open(DFTB_FOLDER + "/detailed.out", "r")
            time_to_read = False
            DFTB_MULLIKEN = np.zeros([structure_DFTB.get_global_number_of_atoms(), 4])
            while True:
                next_line = f2.readline()
                if next_line == " \n":
                    time_to_read = False

                if time_to_read:
                    af = next_line.split()
                    at = int(af[0]) - 1
                    orb = int(af[2])
                    DFTB_MULLIKEN[at][orb] += float(af[3])

                if " Atom Sh.   l       Population" in next_line:
                    time_to_read = True

                if not next_line:
                    break

            # READ NET CHARGES
            f2 = open(DFTB_FOLDER + "/detailed.out", "r")
            time_to_read = False
            DFTB_CHARGES = np.zeros([structure_DFTB.get_global_number_of_atoms(), 1])
            while True:
                next_line = f2.readline()
                if next_line == " \n":
                    time_to_read = False

                if time_to_read:
                    af = next_line.split()
                    at = int(af[0]) - 1
                    DFTB_CHARGES[at, 0] = float(af[1])

                if " Atom           Charge" in next_line:
                    time_to_read = True

                if not next_line:
                    break

            f2.close

            calculator = SinglePointCalculator(
                structure_DFTB, energy=EDFTB, free_energy=EDFTB, forces=DFTB_forces
            )

            structure_DFTB.calc = calculator
            structure_DFTB.get_potential_energy()
            structure_DFTB.get_forces()
            DFTB_DB.write(
                structure_DFTB,
                DFTB=True,
                key=counter,
                data={"fnet_charges": DFTB_CHARGES, "fnet_mull": DFTB_MULLIKEN},
            )
            fnet_strucs.append(structure_DFT)
            fnet_features.append(DFTB_MULLIKEN)
            fnet_energies[counter - 1, 0] = EDFT - EDFTB

    f.close()
    if Fortnet == True:
        fnetdata = Fnetdata(
            atoms=fnet_strucs,
            targets=np.asarray(fnet_energies),
            features=fnet_features,
            atomic=False,
        )
        fnetdata.dump("fnetdata.hdf5")


def main():
    print("--- USAGE:  ccs_build_db MODE [...] --- ")
    print(" ")
    print("       The following modes and inputs are supported:")
    print("       CCS:  file_list(string) DFT.db(string) greedy(bool)")
    print("       DFTB: file_list(string) DFT.db(string) DFTB.db(string) Fortnet(bool)")
    print(" ")

    mode = sys.argv[1]
    file_list = sys.argv[2]
    DFT_data = sys.argv[3]
    print("    Mode: ", mode)
    if mode == "CCS":
        greedy = bool(sys.argv[4])
        print("    DFT data base: ", DFT_data)
        print("    Greedy mode: ", greedy)
        print("")
        print("-------------------------------------------------")
        ccs_build_db(mode, DFT_DB=DFT_data, file_list=file_list, greedy=greedy)
    if mode == "DFTB":
        DFTB_data = sys.argv[4]
        Fortnet = bool(sys.argv[5])
        print("    DFT data base: ", DFT_data)
        print("    DFTB data base: ", DFTB_data)
        print("    Create Fortnet data: ", Fortnet)
        print("")
        print("-------------------------------------------------")
        ccs_build_db(
            mode,
            DFT_DB=DFT_data,
            DFTB_DB=DFTB_data,
            file_list=file_list,
            Fortnet=Fortnet,
        )


if __name__ == "__main__":
    main()
